/***********************************************************************\
*                               runtime.d                               *
*                                                                       *
*                         Unlambda to D compiler                        *
*                                                                       *
*                              Version 0.1                              *
*                                                                       *
*                            Runtime support                            *
>***********************************************************************<
* Copyright (C) 2006-2007 Stewart Gordon.                               *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\***********************************************************************/
module smjg.prog.unlambda.runtime;
import
	std.cstream,
	std.ctype,
	std.stdio,
	std.c.stdlib;

/**
 *	Abstract base class from which all function classes are derived.
 */
abstract class Function {
	/**
	 *	Applies this function to the given expression.
	 *
	 *	Params:
	 *		exp - expression to apply this to.
	 *
	 *	Returns:
	 *		the result of the application.
	 */
	abstract Function opCall(Function delegate()[1] exp...);
}


/**
 *	Abstract base class for all functions that eagerly evaluate
 *	their arguments, i.e. all except for the d builtin.
 */
abstract class NormalFunction : Function {
	/**
	 *	Evaluates the given expression, and then applies this function to it.
	 *
	 *	Params:
	 *		exp - expression to apply this function to.
	 *
	 *	Returns:
	 *		apply(f), where f is the result of evaluating exp.
	 */
	final Function opCall(Function delegate()[1] exp...) {
		return apply(exp[0]());
	}

	/**
	 *	Applies this function to another function.
	 *
	 *	Params:
	 *		f - function to apply this function to.
	 *
	 *	Returns:
	 *		the result of the application.
	 */
	abstract Function apply(Function f);
}


/**
 *	The s builtin: substitution generator.
 *
 *	```sABC evaluates to ``AC`BC.
 */
class SFunction : NormalFunction {
	/**
	 *	Constructs the first partial application of s from the given
	 *	function.
	 *
	 *	Params:
	 *		f - the first function to which s is applied.
	 *
	 *	Returns:
	 *		an SFunction1 representing the partial application.
	 */
	Function apply(Function f) {
		return new SFunction1(f);
	}

	char[] toString() {
		return "S";
	}
}

/**
 *	The s builtin applied to one argument.  That is, a function of the form
 *	`sA.
 */
class SFunction1 : NormalFunction {
	Function c1;

	this(Function a1) {
		c1 = a1;
	}

	/**
	 *	Constructs the second partial application of s from the given
	 *	function.
	 *
	 *	Params:
	 *		f - the second function to which s is applied.
	 *
	 *	Returns:
	 *		an SFunction2 representing the partial application.
	 */
	Function apply(Function f) {
		return new SFunction2(c1, f);
	}

	char[] toString() {
		return "S[" ~ c1.toString ~ "]";
	}
}

/**
 *	The s builtin applied to two arguments.  That is, a function of the form
 *	``sAB.  This is a complete substituted application.  When it is applied
 *	to one more argument C, the expression ``AC`BC is evaluated.
 */
class SFunction2 : NormalFunction {
	Function c1;
	Function c2;

	this(Function a1, Function a2) {
		c1 = a1;
		c2 = a2;
	}

	/**
	 *	Applies the substitution to the given argument.
	 *
	 *	Params:
	 *		f - the third function to which s is applied.
	 *
	 *	Returns:
	 *		the evaluation of the substituted application.
	 */
	Function apply(Function f) {
		return c1(f)(c2(f));
	}

	char[] toString() {
		return "S[" ~ c1.toString ~ ", " ~ c2.toString ~ "]";
	}
}


/**
 *	The k builtin: constant function generator.
 *
 *	``kAB evaluates to A.
 */
class KFunction : NormalFunction {
	/**
	 *	Constructs the first partial application of k from the given
	 *	function.
	 *
	 *	Params:
	 *		f - the first function to which k is applied.
	 *
	 *	Returns:
	 *		a KFunction1 representing the partial application.
	 */
	Function apply(Function f) {
		return new KFunction1(f);
	}

	char[] toString() {
		return "K";
	}
}

/**
 *	The k builtin applied to one argument.  That is, a function of the form
 *	`kA.  This is a constant function.  When it is applied to one more
 *	argument, this argument is discarded and the constant of the function is
 *	returned.
 */
class KFunction1 : NormalFunction {
	Function c1;

	this(Function a1) {
		c1 = a1;
	}

	/**
	 *	Applies the constant function to the given argument.
	 *
	 *	Params:
	 *		f - the second argument to which k is applied.
	 *
	 *	Returns:
	 *		the function with which this KFunction1 was constructed.
	 */
	Function apply(Function f) {
		return c1;
	}

	char[] toString() {
		return "K[" ~ c1.toString ~ "]";
	}
}


/**
 *	The i builtin: identity function.
 *
 *	`iA evaulates to A.
 */
class IFunction : NormalFunction {
	/**
	 *	Applies the identity function to the given argument.
	 *
	 *	Params:
	 *		f - the argument to which i is applied.
	 *
	 *	Returns:
	 *		f
	 */
	Function apply(Function f) {
		return f;
	}

	char[] toString() {
		return "I";
	}
}


/**
 *	The v builtin: black hole.
 *
 *	`vA evaluates to v.
 */
class VFunction : NormalFunction {
	/**
	 *	Applies the black hole function to the given argument.
	 *
	 *	Params:
	 *		f - the argument to which v is applied.
	 *
	 *	Returns:
	 *		v
	 */
	Function apply(Function f) {
		return this;
	}

	char[] toString() {
		return "V";
	}
}


/**
 *	The c builtin: call with current continuation.
 *
 *	In essence, the continuation is the state of the function call stack at
 *	the time c is invoked.  In essence, it's similar to a try/catch
 *	mechanism.  See the Continuation class for more details.
 *
 *	Bug:
 *		The continuation can escape from the c call.  According to the
 *		Unlambda spec, applying a continuation still works in such a case,
 *		by winding the call stack back up.  However, this will not work at
 *		the moment.  I believe that it can be done, if only setjmp can be
 *		made to work in D.
 */
class CFunction : NormalFunction {
	/**
	 *	Applies the given argument to the current continuation.
	 *
	 *	Params:
	 *		f - the function to call on the current continuation.
	 *
	 *	Returns:
	 *		the result of evaluating `f<cont>.
	 */
	Function apply(Function f) {
		Continuation cont = new Continuation;
		try {
			return f(cont);
		} catch (Continuation ex) {
			if (ex is cont) {
				return ex.value;
			} else {
				throw cont;
			}
		}
	}

	char[] toString() {
		return "C";
	}
}


/**
 *	The continuation generated by the c builtin when it is applied.
 *
 *	In essence, the continuation is the state of the function call stack at
 *	the time c is invoked.
 *
 *	If the continuation is applied, then the c function immediately exits,
 *	returning the continuation's argument.
 */
class Continuation : NormalFunction {
	Function value;

	/**
	 *	Causes the c function instance that created this continuation to
	 *	terminate immediately, returning the given argument.
	 *
	 *	Params:
	 *		f - the function to return as the result of calling c.
	 */
	Function apply(Function f) {
		value = f;
		throw this;
	}
}


/**
 *	The d builtin: delay evaluation.
 *
 *	Unlike all other functions, an expression to which d is applied is not
 *	immediately evaluated.  Rather, the expression itself is saved for future
 *	evaluation.
 *
 *	To evaluate ``dAB, first B is evaluated, then A is evaluated.  The result
 *	is then the evaluation of A applied to the evaluation of B.
 */
class DFunction : Function {
	Function opCall(Function delegate()[1] exp...) {
		return new DFunction1(exp[0]);
	}

	char[] toString() {
		return "D";
	}
}


/**
 *	The d builtin applied to an expression.  This is referred to in the
 *	Unlambda manual as a "promise to evaluate".
 */
class DFunction1 : NormalFunction {
	Function delegate() d1;

	private this(Function delegate() d) {
		d1 = d;
	}

	Function apply(Function f) {
		return d1()(f);
	}

	char[] toString() {
		return "D[...]";
	}
}


/**
 *	The .x builtin, where x is any character.  This behaves like the identity
 *	function with the side effect of writing the character x to the standard
 *	output.
 */
class OutCharFunction : NormalFunction {
	char ch;

	private this(char c) {
		ch = c;
	}

	static OutCharFunction opIndex(char c) {
		OutCharFunction* result = c in OC;

		if (result == null) {
			return OC[c] = new OutCharFunction(c);
		}
		return *result;
	}

	Function apply(Function f) {
		dout.write(ch);
		return f;
	}

	char[] toString() {
		return "." ~ ch;
	}
}


char charBuf;


/**
 *	The @ builtin.  This retrieves a character from the standard input and
 *	stores it in the internal buffer.
 *
 *	`@A evaluates to `Ai if the input operation was successful, or
 *	`Av if unsuccessful (e.g. end of file).
 */
class InBufFunction : NormalFunction {
	Function apply(Function f) {
		dout.flush();
		charBuf = din.getc;
		return f(charBuf == char.init ? V : I);
	}

	char[] toString() {
		return "@";
	}
}


/**
 *	The | builtin.  This converts an input character into a .x function.
 *
 *	`|A evaluates to `A.x, where x is the character in the internal buffer,
 *	or `Av if no character is in the buffer.
 */
class ReprintBufFunction : NormalFunction {
	Function apply(Function f) {
		return f(charBuf == char.init ? V : OutCharFunction[charBuf]);
	}

	char[] toString() {
		return "|";
	}
}


/**
 *	The ?x builtin, where x is any character.  This compares x with the
 *	character in the internal buffer.
 *
 *	`?xA evaluates to `Ai if the characters match, or `Av if they don't.
 */
class CompBufFunction : NormalFunction {
	char ch;

	this(char c) {
		ch = c;
	}

	static CompBufFunction opIndex(char c) {
		CompBufFunction* result = c in CB;

		if (result == null) {
			CB[c] = *result = new CompBufFunction(c);
		}
		return *result;
	}

	Function apply(Function f) {
		return f(ch == charBuf ? I : V);
	}

	char[] toString() {
		return "?" ~ ch;
	}
}


/**
 *	The e builtin.  This immediately exits the program.
 */
class EFunction : NormalFunction {
	Function apply(Function f) {
		exit(0);
		assert (false);
	}
}


/**
 *	Unlambda builtins.
 */
Function S, K, I, V, C, D, IB, RB, E, R/+, CR+/;

/**
 *	A few ready-made representations of common expressions.
 */
Function KI, KK, KS, SI, SK, SS, SII;

OutCharFunction[char] OC;
CompBufFunction[char] CB;

static this() {
	S = new SFunction;
	K = new KFunction;
	I = new IFunction;
	V = new VFunction;
	C = new CFunction;
	D = new DFunction;
	IB = new InBufFunction;
	RB = new ReprintBufFunction;
	E = new EFunction;
	R = OutCharFunction['\n'];
	//CR = CompBufFunction['\n'];
	SS = new SFunction1(S);
	/*	`sk is equivalent to `ki in the absence of side effects, but they
	 *	differ in that ```skXY evaluates `XY, but `ki doesn't.
	 */
	SK = new SFunction1(K);
	SI = new SFunction1(I);
	KS = new KFunction1(S);
	KK = new KFunction1(K);
	KI = new KFunction1(I);
	SII = new SFunction2(I, I);
}
